package org.simantics.modeling.ui.componentTypeEditor;

import java.io.PrintWriter;
import java.io.StringWriter;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.AnnotationModel;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.ui.texteditor.AbstractDocumentProvider;
import org.simantics.Simantics;
import org.simantics.databoard.Bindings;
import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.WriteGraph;
import org.simantics.db.common.request.UniqueRead;
import org.simantics.db.common.request.WriteRequest;
import org.simantics.db.exception.DatabaseException;
import org.simantics.db.procedure.Listener;
import org.simantics.db.request.Read;
import org.simantics.modeling.ComponentTypeScriptRequest;
import org.simantics.modeling.ComponentTypeScriptResult;
import org.simantics.scl.compiler.errors.CompilationError;
import org.simantics.scl.compiler.errors.Locations;
import org.simantics.structural.stubs.StructuralResource2;
import org.simantics.structural2.utils.StructuralUtils;
import org.simantics.ui.workbench.ResourceEditorInput;
import org.simantics.utils.logging.TimeLogger;

public class ComponentTypeScriptDocumentProvider extends AbstractDocumentProvider {

    protected ComponentTypeScriptEditor editor;
    
    protected Resource resource;
    protected String currentText;
    protected boolean immutable;
    protected boolean errorHappened;
    
    protected AnnotationModel annotationModel = new AnnotationModel();
    
    protected boolean annotationsInitialized = false;
    
    public ComponentTypeScriptDocumentProvider(ComponentTypeScriptEditor editor) {
        this.editor = editor;
    }

    @Override
    protected IDocument createDocument(Object element) throws CoreException {
        ResourceEditorInput input = (ResourceEditorInput)element;
        resource = input.getResource();
        try {
            return Simantics.getSession().syncRequest(new UniqueRead<Document>() {
                @Override
                public Document perform(ReadGraph graph) throws DatabaseException {
                    StructuralResource2 STR = StructuralResource2.getInstance(graph);
                    currentText = graph.getRelatedValue(resource, STR.ComponentTypeScript_code, Bindings.STRING);
                    Resource owner = graph.getPossibleObject(resource, STR.ComponentType_hasScript_Inverse);
                    immutable = owner != null && StructuralUtils.isImmutable(graph, owner);
                    errorHappened = false;
                    return new Document(currentText != null ? currentText : ""); //$NON-NLS-1$
                }
            });
        } catch (DatabaseException e) {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            errorHappened = true;
            return new Document(sw.toString());
        }
    }
    
    protected void updateAnnotations() {
        Simantics.getSession().asyncRequest(new Read<ComponentTypeScriptResult>() {
            @Override
            public ComponentTypeScriptResult perform(ReadGraph graph) throws DatabaseException {
                // is this the correct way to obtain the parent?
                StructuralResource2 str = StructuralResource2.getInstance(graph);
                Resource componentType = graph.getSingleObject(resource, str.ComponentType_hasScript_Inverse);
                
                return graph.syncRequest(new ComponentTypeScriptRequest(resource, componentType));
            }
        }, new Listener<ComponentTypeScriptResult>() {

            @Override
            public void execute(ComponentTypeScriptResult result) {
                synchronized(annotationModel.getLockObject()) {
                    annotationModel.removeAllAnnotations();
                    for(CompilationError error : result.getErrors()) {
                        Annotation annotation = new Annotation("org.eclipse.ui.workbench.texteditor.error", true, error.description); //$NON-NLS-1$
                        int begin = Locations.beginOf(error.location);
                        int end = Locations.endOf(error.location);
                        Position position = new Position(begin, end - begin);
                        annotationModel.addAnnotation(annotation, position);
                    }
                }
            }

            @Override
            public void exception(Throwable t) {
                t.printStackTrace();
            }

            @Override
            public boolean isDisposed() {
                return editor.isDisposed();
            }
        });
    }
    
    @Override
    protected void doSaveDocument(IProgressMonitor monitor, Object element,
            IDocument document, boolean overwrite) throws CoreException {
        TimeLogger.resetTimeAndLog(getClass(), "doSaveDocument"); //$NON-NLS-1$
        currentText = document.get();
        Simantics.getSession().asyncRequest(new WriteRequest() {
            @Override
            public void perform(WriteGraph graph) throws DatabaseException {
                graph.markUndoPoint();
                StructuralResource2 STR = StructuralResource2.getInstance(graph);
                graph.claimLiteral(resource, STR.ComponentTypeScript_code, currentText, Bindings.STRING);
            }
        });
    }

    @Override
    protected IAnnotationModel createAnnotationModel(Object element) throws CoreException {
        if(!annotationsInitialized) {
            updateAnnotations();
            annotationsInitialized = true;
        }
        return annotationModel;
    }

    @Override
    protected IRunnableContext getOperationRunner(IProgressMonitor monitor) {
        return null;
    }
    
    @Override
    public boolean isModifiable(Object element) {
        return !errorHappened && !immutable;
    }

    @Override
    public boolean isReadOnly(Object element) {
        return errorHappened || immutable;
    }

}
